;+
; NAME:
;      BOXCARFUNC
;
; PURPOSE:
;      Function for fitting the data with a median or smooth fit.
;
; CALLING SEQUENCE:
;      Result = BOXCARFUNC (XVALS, DATAV, VARV, SPECV, COEFF, BOXCARHW)
;
; INPUTS:
;      XVALS:    The x values for the data
;      DATAV:    The vector with cosmic rays to be fitted
;      VARV:     The variance vector
;      SPECV:    The spectrum vector
;      EVAL:     Set to 1 to evaluate the inputs
;      COEFF:    A filler in this routine
;      BOXCARHW: The halfwidth of the median filter
;
; KEYWORDS:
;      None
;
; OUTPUTS
;      Returns an estimation of the data by returning datav ran
;      through a median filter.  If eval is set the routine smooths
;      over the good pixels, then interpolates the bad pixels. 
;
; MODIFICATION HISTORY:
;       Written by: John Dermody, 10 Oct 2003
;
;       Thu Dec 18 22:53:47 2003, John Dermody
;		cleaned up and documented.  
;
;
;-

function BOXCARFUNC, xvals, datav, varv, specv, eval, coeffv, boxcarhw

if (eval eq 0) then begin
  return, median(datav/specv, boxcarhw*2+1) ; median seems to work the best
endif

gv = where(specv ne 0)
bv = where(specv eq 0)
goodx = xvals[gv]
badx  = xvals[bv]
estg = smooth(datav[goodx]/specv[goodx], boxcarhw*2+1, /edge_truncate)
estb = interpol(estg, goodx, badx)
fiteval = datav/specv
fiteval[goodx] = estg
fiteval[badx ] = estb

return, fiteval

end
