; NAME:
;      EXTRSPEC
;
; PURPOSE:
;      optimally extract spectra using weighted profiles
;
; CALLING SEQUENCE:
;      Result = EXTRSPEC(Dataim, Profim, Varim, Bgim, v0, Q, x1, x2, $
;                        Crmask=crmask, Thresh=thresh, Verbose=verbose
;                        GotoVect=Gotovect, Crim=crim, Skyvarim=skyvarim,
;                        Opvar=opvar, Plottype=plottype)
;
; INPUTS:
;      DATAIM: sky-subtracted, processed image to extract spectrum
;              from.  Horizontal is pixel position, vertical is
;              wavelength, assumes no curvature.
;      PROFIM: image of spatial profiles
;      VARIM:  variance image from processed image
;      BGIM:   sky background image, interpolated across the spectrum
;      V0:     root(v0) is squared readout noise in DN
;      Q:      effective number of photons per DN
;      X1,X2:   boundaries in x which contain spectrum (spectrum is
;               inclusive of X1,X2)
;
; OPTIONAL KEYWORDS:
;      CRMASK:   mask of 1's, 0 where cosmic ray position determined by
;                sigma rejection (default is all 1's)
;      SKYVARIM: variance image of sky frame (smoothed)
;      THRESH:   sigma threshold for rejection as cosmic ray, default 5
;      CRIM:     holds dataim-spec*profim*profim/varim
;      OPVAR:    The optimally extracted spectrum's variance
;      GOTOVECT: Stop at that row
;      PLOTTYPE: The type of plot to output to screen
;      VERBOSE:  the level of output to the screen
;      ERRVECT:  the mask with 0's meaning the data was not able to be
;                extracted at that row
;      BPCT:     percentage of bad pixels that halt extraction
;
; OUTPUTS:
;      returns fopt, an array of optimally extracted spectra, and their
;      variances
;
; PROCEDURE:
;      Horne Step 8
;      denominator = (sum over x) mask*profim*profim / varim
;      fopt = (sum over x) mask*profim*dataim/varim / denominator
;      var = (sum over x) mask*profim
;
; MODIFICATION HISTORY:
;       Written by:      Dara Zeehandelaar, 4 March 2003
;       25 March 2003:   omitted loops in favor of matrix math
;       10 Dec 2003:     JD: added verbosity levels

function extrspec, dataim, profim, varim, bgim, v0, Q, x1, x2, $
                   crmask=crmask, thresh=thresh, crim=crim, $
                   skyvarim=skyvarim, opvar=opvar, gotovect=gotovect, $
                   errvect=errvect, verbose=verbose, bpct=bpct, $
                   plottype=plottype

dims = size(dataim)
xdim = dims[1]
nspec = dims[2]

if not keyword_defined(crmask) then crmask = bytarr(xdim,nspec) + 1
if not keyword_defined(thresh) then thresh = 5
if not keyword_defined(skyvarim) then skyvarim = dblarr(xdim,nspec)

if (verbose gt 1) then print, "Into extraction routine"

crim = dblarr(xdim, nspec)      ; initialize for output
crv = dblarr(xdim)
fopt = dblarr(nspec)
opvar = dblarr(nspec)

func = "extractfunc"

errvect = bytarr(nspec) + 1
for i = 0, nspec-1 do begin
  varv    = varim[x1:x2,i]             ; prepare for modification
  maskv   = crmask[x1:x2,i]
  datav   = dataim[x1:x2,i]
  multv   = profim[x1:x2,i]
  bgv     = bgim[x1:x2,i]
  skyvarv = skyvarim[x1:x2,i]
  if keyword_defined(gotovect) then if (i eq gotovect) then stop
  fopt[i] = procvect(datav, varv=varv, multv=multv, verbose=verbose, $
                   thresh=thresh, maskv=maskv, crv = crv, /extract, $
                   bgv=bgv, Q=Q, v0=v0, skyvarv=skyvarv, errflag=errflag, $
                   opvar=opvarv, vectnum=i, func=func, parm=parm, bpct=bpct, $
                   plottype=plottype)
  if (errflag) then errvect[i] = 0
  opvar[i] = parm
  crmask[x1:x2,i] = maskv
  varim[x1:x2,i] =  varv
  crim[x1:x2,i] =   crv * maskv
endfor

return,fopt

end
