; NAME:
;      FITBG
;
; PURPOSE: 
;      creates a sky background image from a data image,
;      interpolated across the spectrum, to create an image of sky
;      lines only
;
; CALLING SEQUENCE:
;      Result = FITBG(Dataim, X1, X2, Degsky=degsky, Q=Q, v0=v0, $
;                     Crmask=crmask, Varim=varim, Bthresh=bthresh, $
;                     Bgmask=bgmask, Profim=profim, spec=spec, $
;                     Skyvarim=skyvarim, verbose=verbose, plottype=plottype)
;
; INPUTS:
;      DATAIM:  The data image, with spectrum
;      X1,X2:   boundaries in x which contain spectrum (spectrum is
;               inclusive of X1,X2)
;
; OPTIONAL KEYWORDS:
;      DEGSKY:   degree of polynomial interpolation (default is 1)
;      CRMASK:   the main mask used by all functions
;      VARIM:    the array that contains the variance of the locations
;      BTHRESH:  the threshold for sigma rejection
;      BCRIM:    the residuals for cosmic ray rejection
;      Q:        The electron per adu of the data
;      v0:       Rn^2 of data
;      BGMASK:   the output mask of cosmic rays found
;      PROFIM:   Used to subtract off the spectrum if iterating
;      SPEC:     The optimally extracted spectrum
;      SKYVARIM: The variance image for the sky subtraction
;      VERBOSE:  Level of printed output
;      NOBGIM:   Set to not fit a background and assume the average
;      GOTOVECT: The row at which to stop the loop
;      ERRVECT:  Will contain the rows that were recovered
;      RECOVER:  Set to attempt recovery of rows that halted fitting
;      BPCT:     The percentage of bad pixels that kicks out of an iteration
;      PLOTTYPE: The type of plot to show while running
;     
; OUTPUT:
;      an array of size dataim in which, for each wavelength, the
;      spectrum from X1 to X2 has been removed, interpolated over
;      (with coefficients COEFFS) and polynomial has been evaluated at
;      all x values
;      
; PROCEDURE:
;      Horne Step 3
; 
; FUNCTIONS CALLED:
;      polyeval, procvect
;
; MODIFICATION HISTORY:
;       Written by:      Dara Zeehandelaar, 25 March 2003
;       8 April 2003:    finished interpolation routine
;                        added keyword COEFFS, made routine to return skyim
;       15 April 2003:   removed dependence on function concat
;       16 April 2003:   debug
;       6 May 2003:      takes mask as input, put skyproc into
;                        iteration within driver
;       20 May 2003:     changed default degree to 1 (gives best fit)
;       16 Jun 2003:     John Dermody: adding variance image keyword
;                        to allow weighted fits.  Also added basic
;                        sigma rejection scheme
;       8 Jul 2003:      JD: removed fit and rejection to its own
;                        function, and renamed file
;       19 Oct 2003:     JD: Matched header to actual calls
;       18 Dec 2003:     JD: Plots, verbosity, and a bunch of other stuff

function fitbg, dataim, x1, x2, degsky=degsky, crmask=crmask, $
                varim=varim, bthresh=bthresh, bcrim=bcrim, Q=Q, v0=v0, $
                bgmask=bgmask, profim=profim, spec=spec, skyvarim=skyvarim, $
                verbose=verbose, nobgim=nobgim, gotovect=gotovect, $
                errvect=errvect, recover=recover, bpct=bpct, plottype=plottype

if not keyword_defined(degsky) then degsky = 1
if not keyword_defined(bthresh) then bthresh = 3.

dims = size(dataim)
xdim = dims[1]
nspec = dims[2]

if not keyword_defined(crmask) then crmask = bytarr(xdim,nspec) + 1
if not keyword_defined(varim) then varim = dblarr(xdim,nspec) + 1
if not keyword_defined(profim) then profim = dblarr(xdim,nspec)
if not keyword_defined(skyvarim) then skyvarim = dblarr(xdim,nspec)

if keyword_set(nobgim) then begin
    bgim = (dblarr(xdim,nspec) + 1) * $
      median([dataim[0:(x1-1),*], dataim[x2+1:xdim-1,*]])
    return, bgim
endif 

if not keyword_defined(spec) then begin 
  xvals1 = dindgen(x1)              ; if there is no spec, then ignore x1:x2
  xvals2 = dindgen(xdim-1-x2)+x2+1  ; starts at xvalue x2+1, goes to xdim-1
  xvals = [xvals1,xvals2]
  specadj = dblarr(xdim,nspec)
endif else begin
  xvals = dindgen(xdim)           ; use spectrum to estimate over x1:x2
  specadj = (spec ## (dblarr(xdim)+1))*profim
endelse

skyint = dblarr(xdim,nspec)
bcrim = dblarr(xdim,nspec)
bgmask = bytarr(xdim,nspec) + 1
func = "polyfunc"
parm = degsky

errvect = bytarr(nspec) + 1
if (verbose gt 1) then print, "Into fit background:"
skyint = dataim
skyint[x1:x2,*] = 0
for i=0, nspec-1 do begin
    datav = dataim[*,i] - specadj[*,i] ; remove spectrum from data
    maskv = crmask[*,i]
    varv = varim[*,i]
    bcrv = bcrim[*,i]
    skyvarv = skyvarim[*,i]
    if ((plottype eq 2) or (verbose eq 5)) then begin
      !p.multi = [0,2,2,1,1]
      !p.charsize = 1.0
      plot, datav,   title='Data Vector', /ystyle
      plot, maskv,   title='Input Mask'
      plot, varv,    title='Variance', /ystyle
      plot, skyvarv, title='Sky variance', /ystyle
      !p.multi = 0
      !p.charsize = 1.0
      wait, 0.01
    endif 
    if (verbose eq 5) then stop
    if keyword_defined(gotovect) then if (i eq gotovect) then stop
    skyint[*,i] = procvect(datav, varv=varv, maskv=maskv, crv=bcrv, $
                           thresh=bthresh, xvals=xvals, errflag=errflag, $
                           Q=Q, v0=v0, skyvarv=skyvarv, verbose=verbose, $
                           vectnum=i, func=func, parm=parm, bpct=bpct, $
                           plottype=plottype)
    if (errflag) then errvect[i] = 0
    if (plottype eq 1) then begin
      sy1 = (i - 10) > 0
      sy2 = (i + 10) < nspec-1
      shade_surf, skyint[*,sy1:sy2], charsize=2, $
        title="Background after processing # " + strtrim(i, 1)
      wait, 0.01
    endif
    varim[xvals,i] = varv[xvals]
    bgmask[*,i] = maskv
    bcrim[*,i] = bcrv * maskv
endfor

; recover bad vectors
if keyword_defined(gotovect) then begin
  surface, pprofim, charsize=1.5, title="Raw estimated profile", $
    xtitle="x vals", ytitle="y vals"
  stop
endif
if keyword_set(recover) then begin
  skyint = recvect(skyint, errvect, plotttype=plottype)
endif

return,skyint

end
