; NAME:
;      FITPROF
;
; PURPOSE:
;      creates an image of normalized spatial profiles, enforces
;      positivitiy, normalization
;
; CATEGORY:
;      astronomy data analysis: optimal spectral extraction
;
; CALLING SEQUENCE:
;        Result = FITPROF(Dataim, Spec, x1, x2, /Nomooth, Degprof=degprof, $
;                         Varim=varim, mainmask=mainmask, profmask=profmask, $
;                         fitgauss=fitgauss, pthresh=pthresh, bgim=bgim, Q=Q, $
;                         v0=v0, skyvarim=skyvarim, pcrim=pcrim, $
;                         fitboxcar=fitboxcar, boxcarhw=boxcarhw, $
;                         straighten=straighen, pprofim=pprofim, $
;                         verbose=verbose centroid=centroid, $
;                         plottype=plottype, tracedeg=tracedeg
;
; INPUTS:
;      DATAIM:  sky-subtracted processed image, horizontal is pixel, 
;               vertical is wavelength
;      SPEC:    array holding extracted spectra, same dimension as
;               vertical of datatim
;      X1,X2:   boundaries in x which contain spectrum (spectrum is
;               inclusive of X1,X2)
;
; OPTIONAL KEYWORDS:
;      NOSMOOTH:   if set, will not smooth profile image in wavelength
;                  direction
;      DEGPROF:    degree of smoothing for profile, default = 2
;      VARIM:      the variance image for weighting of the poly_fit
;      MAINMASK:   the mask used for all functions
;      PROFMASK:   the mask modified by the fit vector routine
;      PTHRESH:    the threshold for sigma rejection
;      FITGAUSS:   set to fit profile in spacial direction with a
;                  gaussian
;      BGIM:       the background image to subtracted off the data
;      Q:          the gain of the image
;      v0:         the readnoise ^2 of the data
;      SKYVARIM    the cooresponding sky variance image
;      PCRIM:      the final residuals for fitting
;      FITBOXCAR:  set to fit the profile with a boxcar filter
;      BOXCARHW:   the width of the median in boxcar average
;      GOTOVECT:   stop when this vector is reached
;      STRAIGHTEN: set to straighten the profile along the x direction
;      VERBOSE:    level of output to screen
;      RECOVER:    attempt to recover the vectors that halted fitting
;                  by splining against good fits
;      PPROFIM:    The shifted and expanded profile image
;      CENTROID:   Set to use centroid fitting instead
;      ERRVECT:    Will contain the vectors that are recovered
;      PLOTTYPE:   The type of plot to output while running
;      BPCT:       The percentage of allowable bad pixels before
;                  halting
;      TRACEEST:   The estimated pixel shift for each row
;      TRACEDEG:   Degree of fit for estimating shift
;
; OUTPUTS:
;      returns an image, same dimensions as dataim,containing the 
;      spatial profile, (smoothed?), positivity enforced, normalized
;
; PROCEDURE:
;      Horne Step 5
;      divides a reduced data image by a standard-extracted spectrum
;      in lambda
;
; FUNCTIONS CALLED:
;      procvect, gaussfunc
;
; MODIFICATION HISTORY:
;       Written by:      Dara Zeehandelaar, 2 March 2003
;       25 March 2003:   omitted loops in favor of matrix math
;                        removed dimension check (will be done in driver)
;       1 April 2003:    refined matrix math
;       15 April 2003:   added degprof keyword, finished smoothing
;       19 April 2003:   debug normalizing
;       16 June 2003:    John Dermody: inserted weighting and a sigma
;                        clip algorthm inton the smoothing function
;       8 Jul 2003:      JD: removed fit and rejection to its own
;                        function
;       1 Sept 2003:     JD: added spacial fit and cleaning
;       19 Oct 2003:     JD: fixed header
;       03 Dec 2003:     JD: Added recovery
;       18 Dec 2003:     JD: cleaning

function FITPROF, Dataim, Spec, x1, x2, Nosmooth=nosmooth, Degprof=degprof, $
                  Varim=varim, mainmask=mainmask, profmask=profmask, $
                  fitgauss=fitgauss, pthresh=pthresh, bgim=bgim, Q=Q, $
                  v0=v0, skyvarim=skyvarim, pcrim=pcrim, $
                  fitboxcar=fitboxcar, boxcarhw=boxcarhw, gotovect=gotovect, $
                  straighten=straighten, pprofim=pprofim, recover=recover, $
                  verbose=verbose, centroid=centroid, errvect=errvect, $
                  bpct=bpct, plottype=plottype, traceest=traceest, $
                  tracedeg=tracedeg

dims = size(dataim)
xdim = dims[1]
nspec = dims[2]

if not keyword_defined(varim   ) then varim    = dblarr(xdim,nspec) + 1
if not keyword_defined(mainmask) then mainmask = bytarr(xdim,nspec) + 1
if not keyword_defined(pthresh ) then pthresh  = 3
if not keyword_defined(bgim    ) then bgim     = dblarr(xdim,nspec)
if not keyword_defined(skyvarim) then skyvarim = dblarr(xdim,nspec)
if not keyword_defined(degprof ) then degprof = 3
if not keyword_defined(boxcarhw) then boxcarhw = 3
if not keyword_defined(tracedeg) then tracedeg = 4

if (verbose gt 1) then print, "Into fitprof"

if keyword_set(nosmooth) then begin
  ; profile: divide reduced image by standard spectra in lambda
  specim = spec ## (dblarr(xdim) + 1)    
  profim = dataim / specim                 ; divide matrices
  profim = profim * mainmask
  negloc = where(profim lt 0)
  if (negloc ne [-1]) then profim[negloc] = 0
  t = total(profim[x1:x2,*], 1)
  t = t ## (dblarr(xdim) + 1)
  profim = profim / t
  return,profim
endif

; adjust for curved trace
if keyword_set(straighten) then begin
  shiftv = dblarr(nspec)
  xrange = indgen(x2 - x1 + 1) + x1
  mid = x1 / 2 + x2 / 2
  maskv = bytarr(nspec) + 1
  for i = 0, nspec-1 do begin
    datav = dataim[x1:x2, i] 
    varv =  varim [x1:x2, i] 
    specv = spec[i] 
    if keyword_set (centroid) then begin
      cnt = centermass(xrange, datav, varv, specv)
      shiftv[i] = cnt - mid
    endif else begin
      foo = gaussfunc(xrange, datav, varv, specv, 0, coeff)
      if (coeff[0] eq 0) then begin 
        maskv[i] = 0
      endif else begin
        shiftv[i] = coeff[1] - mid
      endelse
    endelse
  endfor
  level = 10
  varv = dblarr(nspec) + (1./level)^2
  func = "polyfunc"
  parm = tracedeg
  if ((plottype eq 2) or (verbose eq 5)) then begin
     !p.multi = [0,2,2,1,1]
     plot, shiftv,  title='Shift Vector', /ystyle
     plot, varv,    title='Variance estimate', /ystyle
     plot, maskv,   title='Input Mask', yrange = [-0.1,1.1]
     !p.multi = 0
     wait, 1.5
  endif
  if (verbose eq 5) then stop
  shiftest = procvect(shiftv, varv=varv, /noupdate, thresh=3, $
                      maskv=maskv, func=func, parm=parm, verbose=verbose, $
                      plottype=plottype)
  if (plottype eq 2) then begin
    plot, shiftv, title = 'Shift in PSF vs Estimate', xtitle= 'Wavelength', $
          ytitle= 'Shift amount in pixels'
    oplot, shiftest
    wait, 1.5
  endif
  traceest = shiftest + mid
  if (verbose eq 5) then stop
endif else begin
  shiftest = dblarr(nspec)
  level = 1
endelse

specim = spec ## (dblarr(xdim) + 1)

pdataim = sampleshift(dataim  [x1:x2,*], level, -shiftest, 1)
pvarim  = sampleshift(varim   [x1:x2,*], level, -shiftest, 1)
pmask   = sampleshift(mainmask[x1:x2,*], level, -shiftest, 1)
pbgim   = sampleshift(bgim    [x1:x2,*], level, -shiftest, 1)
pskyvar = sampleshift(skyvarim[x1:x2,*], level, -shiftest, 1)
pspecim = sampleshift(specim  [x1:x2,*], level, -shiftest, 1)


pnx = (x2 - x1 + 1) * level
nspec = nspec
yvals = dindgen(nspec)
profmask = pmask
pcrim  = dblarr(pnx, nspec)
pspec = total(pdataim, 1)
pprofim = dblarr(pnx, nspec)

; smooth the profile
if keyword_set(fitgauss) then begin
  f1 = 0                 ; fit over spatial dimension
  f2 = nspec - 1
  nvect = pnx
  ndata = nspec
endif else begin
  f1 = 0                 ; fit over spectral dimension
  f2 = pnx - 1
  nvect = nspec
  ndata = pnx
endelse

datav   = dblarr(nvect)  ; initialize so dimensions are correct
maskv   = dblarr(nvect)
varv    = dblarr(nvect)
multv   = dblarr(nvect)
skyvarv = dblarr(nvect)
bgv     = dblarr(nvect)
errvect = dblarr(ndata) + 1

index = dindgen(pnx, nspec)

if keyword_set(fitgauss) then begin
  func = "gaussfunc"
  parm = 1
endif else if keyword_set(fitboxcar) then begin
  func = "boxcarfunc"
  parm = boxcarhw
endif else begin
  func = "polyfunc"
  parm = degprof
endelse

pprofim = pdataim / pspecim
for i = f1, f2 do begin
  if keyword_set(fitgauss) then begin 
    is = index[*,i]
  endif else begin 
    is = index[i,*]
  endelse
  datav[*]   = pdataim[is]
  maskv[*]   = pmask[is]
  varv[*]    = pvarim[is]
  crv        = pcrim[is]
  multv[*]   = pspecim[is]
  skyvarv[*] = pskyvar[is]
  bgv[*]     = pbgim[is]
  if keyword_defined(gotovect) then if (i eq gotovect) then stop
  if ((plottype eq 2) or (verbose eq 5)) then begin
     !p.multi = [0,2,3,1,1]
     !p.charsize = 2.0
     !x.margin = [8, 2]
     !y.margin = [2, 2]
     plot, datav,   title='Data Vector', /ystyle
     plot, multv,   title='Std Spectrum', /ystyle
     plot, maskv,   title='Input Mask', yrange = [-0.1,1.1]
     plot, varv,    title='Variance', /ystyle
     plot, skyvarv, title='Sky variance', /ystyle
     plot, bgv,     title='Background', /ystyle
     !p.multi = 0
     !p.charsize = 1.0
     !x.margin = [10, 3]
     !y.margin = [4, 2]
     wait, 0.01
  endif
  if (verbose eq 5) then stop
  errflag = 0
  pprofim[is] = procvect(datav, varv=varv, thresh=pthresh, multv=multv, $
                         maskv=maskv, crv=crv, bgv=bgv, Q=Q, v0=v0, $
                         skyvarv=skyvarv, verbose=verbose, errflag=errflag, $
                         vectnum=i, func=func, parm=parm, bpct=bpct, $
                         plottype=plottype)
  if (errflag) then begin
    errvect[i] = 0
  endif
  if (plottype eq 1) then begin
    shx1 = ((i - 10) > f1)
    shx2 = ((i + 10) < f2)
    if keyword_set(fitgauss) then begin 
      sis = index[*,shx1:shx2]
    endif else begin 
      sis = index[shx1:shx2,*]
    endelse
    shade_surf, pprofim[sis], charsize=2, $
      title="Profile estimation before # " + strtrim(i, 1)
    wait, 0.01
  endif
  profmask[is] = maskv
  pvarim[is] =   varv
  pcrim[is] =    crv * maskv
endfor

; recover bad vectors
if not keyword_set(fitgauss) then pprofim = transpose(pprofim)
if keyword_defined(gotovect) then begin
  surface, pprofim, charsize=1.5, title="Raw estimated profile", $
    xtitle="x vals", ytitle="y vals"
  stop
endif
if keyword_set(recover) then begin
  pprofim = recvect(pprofim, errvect, plottype=plottype)
endif
if not keyword_set(fitgauss) then pprofim = transpose(pprofim)

; shrink
profim = dblarr(xdim, nspec)
profim[x1:x2,*] = sampleshift(pprofim, 1, shiftest*level, level)
varim[x1:x2,*]  = sampleshift(pvarim , 1, shiftest*level, level)

; enforce positivity
negloc = where(profim lt 0)
if (negloc ne [-1]) then profim[negloc] = 0

; normalize
t = total(profim[x1:x2,*], 1)
t = t ## (dblarr(xdim) + 1)
tprof = profim / t
profim = tprof

return,profim

end
