;
; NAME:
;      GAUSSFUNC
;
; PURPOSE:
;      Function for estimating a Gaussian curve to the data
;
; CALLING SEQUENCE:
;      Result = GAUSSFUNC (XVALS, DATAV, VARV, SPECV, COEFF)
;
; INPUTS:
;      XVALS:  The x values for the data
;      DATAV:  The vector with cosmic rays to be fitted
;      VARV:   The variance vector
;      SPECV:  The spectrum vector
;      EVAL:   Do not fit, instead, evaluate based on coeff
;      COEFF:  Will contain the coefficients of the fit
;      REEST:  Set to reestimate the spectrum after bad pixel rejection
;
; KEYWORDS:
;      None
;
; OUTPUTS
;      Returns an estimation of the data by fitting a Gaussian.  If a
;      good fit is not found quickly to the data, the Gaussian is
;      fitted to the data smoothed instead.  If eval is set it simply
;      evaluates the Gaussian coeffeciants passed in at all points.
;
; MODIFICATION HISTORY:
;      Written by:    John Dermody, 22 Sep 2003
;
;      Thu Dec 18 22:55:25 EST 2003, John Dermody
;            Cleaned up and docummented
;
;-

function GAUSSFUNC, xvals, datav, varv, specv, eval, coeff, reest

if (eval eq 1) then begin
  gausseval, xvals, coeff, ret
  return, ret
endif

hbw = 4                   ; half box width for estimation

if keyword_set(reest) then begin
  specv = int_tabulated(xvals, datav) ; reestimate spectrum
endif

sdata = smooth(datav / specv, 2*hbw+1, /edge_truncate) ; est Gauss

hi = max(sdata)                           ; estimated Gaussian height
center = xvals[median(where(sdata eq hi))]; estimated Gaussian center
fwhm = (size(where(sdata gt hi/2)))[1]    ; estimated Gaussian FullWidthHalfMax
coeff = [hi, center, fwhm/2.354] ; coefficiance for estimate

; so 0 variances can be smoothed
;varv = varv > (min(varv[where(varv gt 0)]) * 1.e-6) 
gl = where(varv ne 0)

est = datav / specv
origcoeff = coeff
est[gl] = curvefit(xvals[gl], datav[gl]/specv, 1./(varv[gl]/specv^2), coeff, $
               iter=fititer, function_name='gausseval', $
               /noderivative, /double)
               ; estimate vector using Gaussian

if (fititer gt 8) then begin  ; estimation wrong so use smooth instead
  est[gl] = curvefit(xvals[gl], sdata[gl]/specv, 1./(varv[gl]/specv^2), $
                     origcoeff, iter=fititer, function_name='gausseval', $
                     /noderivative, /double) ; estimate vector using Gaussian
  coeff = [0, 0, 0]
endif 
est = est / int_tabulated(xvals, est) ; profile is fit / spectrum

return, est
end
