;+ NAME:
;      OPTSPECEXTR
;
; PURPOSE:
;      data pipeline for image processing and optimal extraction
;
; CALLING SEQUENCE:
;        Result = OPTSPECEXTR(Data, Var, RN, Q, X1, X2, Skyvarim=skyvarim, $
;                             Mask=mask, Degsky=degsky, $
;                             Stdspec=stdspec, Varstd=varstd, $
;                             Nosmooth=nosmooth, Degprof=degprof, $
;                             BThresh=bthresh, PThresh=pthresh, $
;                             MThresh=mthresh, Itnum=Itnum, Fitgauss=fitgauss,$
;                             Fitboxcar=fitboxcar, Boxcarhw=boxcarhw, $
;                             Repmain=repmain, Straighten=straighen, $
;                             Varout=varout, opvar=opvar, Crmask=crmask, $
;                             Bgim=bgim, Crim=crim, Profim=profim, $
;                             Profmask=profmask, recover=recover, $
;                             Bgmask=bgmask, verbose=verbose, $
;                             bpct=bpct, tracedeg=tracedeg)
;
; INPUTS:
;      DATA:  processed image (raw-sky)/flat for IR, (raw-bias)/flat
;             for optical (by Horne step 1)
;      VAR:   variance image of data in DN^2 (i.e., Data/Q+Rn^2, Horne step 2)
;      RN:    RN is read noise of data frame in DN
;      Q:     effective number of photons per DN for data frame
;      X1,X2: boundaries in x of spectrum, spectrum is inclusive of X1,X2
;
; OPTIONAL KEYWORDS: (Input)
;      SKYVARIM:   Variance image of sky (by Horne step 2)
;      MASK:       user input mask with good pixels=1, bad=0
;      DEGSKY:     degree of polynomial fit for interpolation in skyproc (1)
;      NOSMOOTH:   will not smooth spatial profile if set
;      DEGPROF:    degree to smooth spatial profile, (default = 3)
;      BTHRESH:    sigma threshold for cosmic ray rejection for bg, (6)
;      PTHRESH:    sigma threshold for profile image, (4)
;      MTHRESH:    sigma threshold for main mask, (5)
;      FITGAUSS:   Set to smooth the profile image using a Gaussian fit
;                  in the spatial direction, instead a polynomial fit in
;                  wavelength direction
;      FITBOXCAR:  Set to smooth the profile using a median filter in
;                  spectral dimension
;      BOXCARHW:   The half-width of the boxcar fit
;      REPMAIN:    Set to iterate the entire procedure until a self
;                  consistant solution is found
;      STRAIGHTEN: Set to straighten the trace, default Gaussian
;      VERBOSE:    Set to level of output: 0 - fatal, 1 - collect
;                  stats, 2 - where you are in the program, 3 - output 
;                  every warning, 4 - plot how the data are being used,
;                  5 - stop after each iteration.
;      PLOTTYPE:   Set to the type of plot outputed to the screen:
;                  1 - use shade_surf to show the fitted and raw data
;                  2 - show the data used as input to procvect
;                  3 - show how procvect is fitting the data
;      CENTROID:   Set to straighten with centroid instead
;      BPCT:       Set the the percentage of bad pixels before kicking out
;      RECOVER:    Set to recover bad vectors in fitting routines
;      MAXREP:     The maximum iterations the driver routine is allowed
;      INTEGRATE:  When calculating stdspec integrate overbad pixels
;      BGOTOVECT:  A debugging tool to halt execution at that vector.
;      PGOTOVECT:  A debugging tool to halt execution at that vector.
;      EGOTOVECT:  A debugging tool to halt execution at that vector.
;      TRACEDEG:   The degree of fitting for estimating shift in trace 
;
; OPTIONAL KEYWORDS: (Output)
;      VAROUT:   returns the final caculated variance
;      OPVAR:   returns optimally extracted variance      
;      STDSPEC:  standard spectrum
;      VARSTD:   returns variance of standard spectrum
;      CRMASK:   calculated cosmic ray mask
;      BGIM:     interpolated sky background image
;      CRIM:     cosmic ray image: dataim-spec*profim*profim/varim
;      PROFIM:   calculated profile image
;      PROFMASK: the calculated profile cosmic ray mask 
;      BGMASK:   The background's mask
;      ITNUM:    number of iterations of Horne procedure
;      BERRVECT: An array with the mask of vectors to attempt recover
;      PERRVECT: An array with the mask of vectors to attempt recover
;      OERRVECT: An array with the mask of vectors to attempt recover
;      TRACEEST: The estimated trace shift for each vector
;
; OUTPUTS:
;      returns FOPT, the optimally extracted spectrum, and OPVAR, the variance
;      
; PROCEDURE:
;      implements Horne optimal extraction algorithm
;
; FUNCTIONS CALLED:     
;      Horne step 3: fitbg
;      Horne step 4: stdexr
;      Horne step 5: fitprof
;      Horne step 6: opvar_ir
;      Horne step 7: cosmask
;      Horne step 8: extrspec
;
; MODIFICATION HISTORY:
;       Written by:      Dara Zeehandelaar, 13 March 2003
;       23 April 2003    took out Horne steps 1 and 2, user does
;                        outside
;       30 April 2003    added optional user mask input
;       6 May 2003:      put skyproc within final iteration in order
;                        to use cosmic ray mask
;       7 May 2003:      added optional keyword edge
;       16 May 2003:     added keywords stdspec, crmask, bgim, crim,
;                        profim, itnum to return intermediate steps if desired
;       10 Aug 2003:     John Dermody: reorganized to have only 3
;                        functions called
;       20 Aug 2003:     JD: added main loop for extreme cases
;       1 Sept 2003:     JD: removed edge and cleaning up.
;       14 Oct 2003:     JD: turned everything into doubles 
;       10 Dec 2003:     JD: set up verbosity levels
;-

function OPTSPECEXTR, data, var, rn, q, x1, x2, skyvarim=skyvarim, $
                      mask=mask, degsky=degsky, stdspec=stdspec, $
                      varstd=varstd, nosmooth=nosmooth, degprof=degprof, $
                      bthresh=bthresh, pthresh=pthresh, mthresh=mthresh, $
                      itnum=itnum, fitgauss=fitgauss, fitboxcar=fitboxcar, $
                      boxcarhw=boxcarhw, repmain=repmain, $
                      straighten=straighten, varout=varout, opvar=opvar, $
                      crmask=crmask, bgim=bgim, crim=crim, profim=profim, $
                      profmask=profmask, pcrim=pcrim, bcrim=bcrim, $
                      bgmask=bgmask, verbose=verbose, nobgim=nobgim, $
                      ogotovect=ogotovect, pgotovect=pgotovect, $
                      bgotovect=bgotovect, bpct=bpct, $
                      integrate=integrate, centroid=centroid, $
                      maxrep=maxrep, recover=recover, $
                      berrvect=berrvect, perrvect=perrvect, $
                      eerrvect=eerrvect, plottype=plottype, $
                      traceest=traceest, tracedeg=tracedeg

nx = (size(data))[1]
ny = (size(data))[2]
if not keyword_defined(mask) then mask = bytarr(nx, ny) + 1
if not keyword_defined(verbose) then verbose = 0
if not keyword_defined(plottype) then plottype = 0
if (verbose lt 4) then plottype = 0

; read noise --> V0 (variances)
v0 = RN * RN
x1 = round(x1)
x2 = round(x2)
crmask = mask
varim = var

  dataim = data

  ; Horne Step 3 - fit sky background
  saveverbose = verbose
  if (verbose eq 5) then stop
  bgim = fitbg(dataim, x1, x2, degsky=degsky, Q=Q, v0=v0, crmask=crmask, $
               varim=varim, bthresh=bthresh, bcrim=bcrim, errvect=Berrvect, $
               bgmask=bgmask, profim=profim, spec=spec, skyvarim=skyvarim, $
               verbose=verbose, nobgim=nobgim, gotovect=Bgotovect, $
               recover=recover, bpct=bpct, plottype=plottype)
  verbose = saveverbose


; to get a more accurate sky, subtract out the spec and refit
if keyword_set(repmain) then done = 0 else done = 1
itnum = 0
repeat begin

  ; copy and use this
  dataim = data
  savemask = crmask
  crmask = mask
  if keyword_defined(spec) then savespec = spec

  ; sky-subtract
  dataim = dataim - bgim

  if not keyword_defined(spec) then begin
    ; Horne Step 4 - extract standard spectrum, compute variance
      spec = stdextr(dataim, varim, x1, x2, crmask, var=varstd, $
                     integrate=integrate)
    stdspec = spec   ; saving for optional output
  endif

  ; Horne Step 5 - construct spatial profile
  ; WITH STEP 6 - revise variance estimates
  saveverbose = verbose
  if (verbose eq 5) then stop
  profim = fitprof(dataim, spec, x1, x2, nosmooth = nosmooth, $
                   degprof = degprof, varim = varim, mainmask = crmask, $
                   profmask = profmask, pthresh = pthresh, bgim = bgim, $
                   Q = Q, v0 = v0, skyvarim = skyvarim, pcrim=pcrim, $
                   fitgauss = fitgauss, fitboxcar = fitboxcar, $
                   boxcarhw = boxcarhw, straighten = straighten, $
                   gotoVect = pgotovect, verbose=verbose, $
                   centroid = centroid, errvect=Perrvect, $
                   recover = recover, bpct=bpct, plottype=plottype, $
                   traceest = traceest, tracedeg=tracedeg)
  verbose = saveverbose

  ; Horne Steps 6, 7, 8 - Optimally Extract Spectrum
  ; Also revise variance estimates and mask cosmic rays
  spec = extrspec(dataim, profim, varim, bgim, v0, Q, x1, x2, $
                  crmask=crmask, opvar=opvar, thresh=mthresh, crim=crim, $
                  skyvarim=skyvarim, verbose=verbose, gotovect=ogotovect, $
                  errvect=Eerrvect, bpct=bpct, plottype=plottype)
  

  if (array_equal(crmask, savemask)) then done = 1
  if keyword_defined(maxrep) then if (itnum ge maxrep) then done = 1
  itnum  = itnum + 1
endrep until (done eq 1)

varout = varim

brv = strjoin(strtrim(where(Berrvect eq 0), 1), " ")
prv = strjoin(strtrim(where(Perrvect eq 0), 1), " ")
erv = strjoin(strtrim(where(Eerrvect eq 0), 1), " ")

if keyword_set(recover) then begin
  initf = "Tried to recover at "
endif else begin
  initf = "Eliminated half the pixels at "
endelse
fsummary = initf
esummary = ""
if (brv ne "-1") then fsummary = fsummary + "rows " + brv + " in BG fit, "
if (prv ne "-1") then fsummary = fsummary + "cols " + prv + " in Prof fit. " 
if (erv ne "-1") then esummary = "Unable to extract spectrum at rows " + erv

if (fsummary eq initf) then begin
  summary = esummary
endif else begin
  summary = fsummary + esummary
endelse
if ((summary ne "") and (verbose gt 0)) then print, summary

return,spec

end
