; NAME:
;      PROCVECT
;
; PURPOSE:
;      Fits a polynomial to a vector whose cosmic rays have been found
;      using sigma rejection
;
; CALLING SEQUENCE:
;      Result = PROCVECT(Datav, Varv=Varv, Multv=multv, Maskv=Maskv, $
;                       Thresh=thresh, Bgv=bgv, Q=Q, v0=v0, deg=deg, $
;                       Skyvarv=skyvarv, Extract=extract, xvals=xvals, $
;                       Fitgauss=fitgauss, Fitboxcar=fitboxcar, $
;                       Boxcarhw=boxcarhw, Noupdate=noupdate, Crv=crv, $
;                       Verbose=verbose, Opvar=opvar, Errflag=errflag,$
;                       plottype=plottype)
;
; INPUTS:
;      DATAV:   The vector with cosmic rays to be fitted
;
; OPTIONAL KEYWORDS:
;      VARV:      The variance vector
;      MULTV:     An array of same length as datav containing values to
;                 be multiplied by the fitted data in the cosmic ray rejection
;      MASKV:     cosmic ray mask (0 for cosmic ray, otherwise 1)
;      THRESH:    sigma threshold for rejection as cosmic ray, default 5
;      XVALS:     Indicies of the data to fit
;      BGV:       An array of wavelength dimension that contains the
;                 background values at datav
;      Q:         The gain of the array
;      V0:        The readnoise of the array
;      SKYVARV:   The sky variance vector
;      CRV:       The residual of the data fit 
;      EXTRACT:   Instead of fitting the data, extract the optimal
;                 spectrum
;      NOUPDATE:  Don't update the variance vector
;      VERBOSE:   Set to 3 to halt the function after every iteration
;      OPVAR:     The optimally extracted spectrum's variance
;      ERRFLAG:   Set if a sutible fit is not found
;      FUNC:      The function to fit to the data.  The first four be xvals,
;                 datav, varv, multv, eval.  It should return an estimation
;      PARM:      The additional paramaters of FUNC.
;      BPCT:      The percentage of bad pixels that will end fitting
;      PLOTTYPE:  Set to 3 to plot each fit itteration result as it runs
;
; OUTPUTS:
;      Returns the output of func after cosmic rays have been eliminated
;
; PROCEDURE:
;      uses sigma rejection to determine positions of cosmic rays,
;      then return polynomial evaluated
;
; MODIFICATION HISTORY:
;      Written by:    John Dermody, 25 Jun 2003
;      15 Jul 2003:   added optimally exctacting spectrum option
;      10 Aug 2003:   allows Gaussian fit
;      20 Aug 2003:   allows RSA fit
;      1 Sept 2003:   cleaning and commenting
;      19 Oct 2003:   cleaning
;      18 Dec 2003:   moved out functions into passing routines and cleaning

function procvect, datav, varv=varv, multv=multv, maskv=maskv, $
                   thresh=thresh, xvals=xvals, bgv=bgv, Q=Q, v0=v0, $
                   skyvarv=skyvarv, crv=crv, extract=extract, $
                   noupdate=noupdate, verbose=verbose, $
                   opvar=opvar, errflag=errflag, vectnum=vectnum, $
                   func=func, parm=parm, bpct = bpct, plottype = plottype

nx = (size(datav))[1]           ; size of vector

if not keyword_defined(thresh  ) then thresh   = 5
if not keyword_defined(deg     ) then deg      = 1
if not keyword_defined(crv     ) then crv      = dblarr(nx)
if not keyword_defined(maskv   ) then maskv    = bytarr(nx) + 1
if not keyword_defined(varv    ) then varv     = dblarr(nx) + 1
if not keyword_defined(xvals   ) then xvals    = dindgen(nx)
if not keyword_defined(multv   ) then multv    = dblarr(nx) + 1.
if not keyword_defined(bgv     ) then bgv      = dblarr(nx)
if not keyword_defined(Q       ) then Q        = 1.
if not keyword_defined(v0      ) then v0       = 0.
if not keyword_defined(skyvarv ) then skyvarv  = dblarr(nx)
if not keyword_defined(bpct    ) then bpct     = 0.5

if keyword_defined(vectnum) then begin
    vectnum_s = " at Vector # " + strtrim(vectnum,1)
endif else begin
    vectnum_s = ""
endelse

errflag = 0
vthresh = thresh * thresh       ; threshold in variance
coeffv  = dblarr(deg + 1)

funcdone = 1                    ; initialize for while loop
funccount = 0                   ; number of loops

while (funcdone eq 1) do begin
  funcdone = 0
  goodvals = where(maskv[xvals] eq 1)
  if (n_elements(goodvals) lt (n_elements(xvals) * bpct)) then begin
    errflag = 1
    fiteval = call_function(func, dindgen(nx), datav, varv, $
                            multv*maskv, 1, coeffv, parm)
    fiteval[*] = 0
    if (verbose gt 2) then print, "Too many pixels rejected" + vectnum_s
    return, fiteval
  endif
  fitx = xvals[goodvals]        ; xvals for good pixels
  fitdata = datav[fitx]         ; data for good pixels
  fitvar = varv[fitx]           ; variance of good pixels
  fitmult = multv[fitx]         ; multiplier for good pixels
  est = call_function(func, fitx, fitdata, fitvar, fitmult, 0, coeffv, parm) 
  crv[fitx] = (fitdata - fitmult * est)^2 / (fitvar > 1.e-6) ; no divide by 0
  badpix = where (crv[fitx] gt vthresh) ; get bad locations
  if ((plottype eq 3)  or (verbose eq 5)) then begin
    savemulti = !p.multi
    !p.multi = [0,1,2,1,1]
    plot, fitx, fitdata/fitmult, $
          title='Actual vs. Fitted' + vectnum_s, $
          ytitle='Data Values (if applicable / Spec)', $
          xtitle='Pixel Locations', charsize=1.1
    if not keyword_set (extract) then oplot, fitx, est
    plot, fitx, sqrt(crv[fitx]), title='Residuals', $
          ytitle='Sigma Difference of Data vs. Expected', $
          xtitle='Pixel Locations', charsize=1.1
    !p.multi = savemulti
    wait, 0.01
  endif
  if (verbose eq 5) then stop
  if (badpix ne [-1]) then begin
    badx = fitx[badpix]
    maxpos = where(crv[badx] eq  max(crv[badx])) ; only elimate max pixel
    maxx = badx[maxpos]
    funccount = funccount + n_elements(maxx) ; add count for bad pix
    maskv[maxx] = 0             ; mask bad pix
    funcdone = 1                ; set so sequence loops
  endif
  if not keyword_set(noupdate) then begin
    varv[fitx] = (abs(fitmult * est + bgv[fitx])) / Q + v0 + skyvarv[fitx]
  endif
endwhile

fiteval = call_function(func, dindgen(nx), datav, varv, $
                        multv*maskv, 1, coeffv, parm)

if not keyword_set(noupdate) then begin
  varv = (abs(multv * fiteval + bgv)) / Q + v0 + skyvarv 
  ; get var for all pixels
end

return, fiteval

end
