; NAME:
;      STDEXTR
;
; PURPOSE: 
;      standard box extraction of spectrum, returns spectrum and
;      variance image
;
; CALLING SEQUENCE:
;      Result = STDEXTR(Dataim,Varim,X1,X2,Crmask,Var=var)
;
; INPUTS:
;      DATAIM:  sky-subtracted, processed image to extract spectrum
;               from.  Horizontal is pixel position, vertical is 
;               wavelength, assumes no curvature.
;      VARIM:   variance image from processed image
;      X1,X2:   boundaries in x over which spectrum will be summed
;      CRMASK:  the cosmic ray mask for the data image
;
; OPTIONAL KEYWORDS:
;      VAR:       variance of extracted spectra, length n where n is the
;                 number of wavelengths taken of the image (vertical direction)
;      INTEGRATE: set to integrate the spectrum instead mearly
;                 addition.  Only useful if input mask has bad pixels
;
; OUTPUT:
;      returns fstd, an array of length n containing extracted spectra,
;      where n is the number of wavelengths of the image (vertical
;      direction)
;
; PROCEDURE:
;      Horne Step 4 (standard box extraction)
;
; MODIFICATION HISTORY:
;       Written by:      Dara Zeehandelaar, 2 March 2003
;       25 March 2003:   removed loops in favor of matrix math
;
;            Thu Dec 18 23:05:16 EST 2003, John Dermody
;                  Cleaned up, added the integration.
; 

function stdextr, dataim, varim, x1, x2, crmask, var=var, integrate=integrate

ny = (size(dataim))[2]

if keyword_set(integrate) then begin
  fstd = dblarr(ny)
  var  = dblarr(ny)
  for i = 0, ny-1 do begin
    xvals = where(crmask[x1:x2, i] eq 1)
    fstd[i] = int_tabulated(xvals, dataim[xvals+x1, i])
    var[i]  = int_tabulated(xvals, varim [xvals+x1, i])
  endfor
endif else begin
  fstd  = total((dataim * crmask)[x1:x2, *], 1)
  var   = total((varim  * crmask)[x1:x2, *], 1)
endelse

return,fstd

end
